#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""A very limited registrant-ep for Resorce Directory plugtests.

It largely utilizes the client registration features already built into
aiocoap, but also offers a --simple optiion to run the simple registration
that's otherwise not needed in aiocoap as the clients are featureful enough to
be able to do (and need features of) full registration.

Even though aiocoap typically does not emit such links, it produces the
spurious links from the plug test description to see whether the RD deals with
them well.

All three site descriptions from the plug test are selectable with --site; it
is up to the operator to pick the right one for the right test, and to decide
whether to register simply, regularly, or using a CT (which is easily simulated
by making the registrant simply-register at a rd-relay). The --register-as
parameter should have a usable default for the plug test.
"""

import argparse
from urllib.parse import urljoin

from aiocoap import Message, POST
from aiocoap.cli.common import (add_server_arguments,
        server_context_from_arguments)
from aiocoap.util.cli import AsyncCLIDaemon
from aiocoap.resource import Resource, Site, WKCResource
from aiocoap.util.linkformat import LinkFormat, Link
from aiocoap.resourcedirectory.client.register import Registerer

class DemoResource(Resource):
    async def render_get(self, request):
        if request.opt.accept == 41:
            return Message(payload=("<data>%s</data>" % self.text).encode('utf8'), content_format=41)
        else:
            return Message(payload=self.text.encode('utf8'), content_format=0)

class Temp(DemoResource):
    rt = "temperature-c"

    text = "39.1°C"

class Light(DemoResource):
    rt = "light-lux"

    text = "There are four lights."

class Light3(DemoResource):
    rt = "light"
    ct = 0

def build_site_node1():
    site = Site()
    temp = Temp()
    site.add_resource(['sensors', 'temp'], temp)
    light = Light()
    site.add_resource(['sensors', 'light'], light)
    for x in (temp, light):
        x.if_ = "sensor"
        x.ct = 41

    def get_links():
        links = site.get_resources_as_linkheader()
        for l in links.links:
            if l.href == "/sensors/light":
                l.attr_pairs.append(("anchor", "coap://spurious.example.com:5683"))
        return LinkFormat(links.links)
    site.add_resource(['.well-known', 'core'], WKCResource(get_links))

    return site

def build_site_node2():
    site = Site()
    temp = Temp()
    site.add_resource(['temp'], temp)
    light = Light()
    site.add_resource(['light'], light)
    for x in (temp, light):
        x.ct = 0

    def get_links():
        links = site.get_resources_as_linkheader()
        links.links.append(Link("/t", anchor="sensors/temp", rel="alternate"))
        links.links.append(Link("http://www.example.com/sensors/t123", anchor="sensors/temp", rel="describedby"))
        return LinkFormat(links.links)
    site.add_resource(['.well-known', 'core'], WKCResource(get_links))

    return site

def build_site_node3():
    site = Site()
    for x in 'left', 'middle', 'right':
        site.add_resource(['light', x], Light3())
    site.add_resource(['.well-known', 'core'], WKCResource(site.get_resources_as_linkheader))
    return site

class RDRegistrant(AsyncCLIDaemon):
    registerer = None
    async def start(self):
        p = argparse.ArgumentParser()
        p.add_argument("rd_uri", help="Preconfigured address of the resource"
                " directory", nargs='?', default='coap://[ff05::fd]')
        p.add_argument("--simple", help="Run simple registration rather than"
                " full (incomplete: Never renews)", action="store_true")
        p.add_argument("--register-as", help="Endpoint name to register as (default: node$SITE",
                default=None)
        p.add_argument("--site", help="Use a different resource / link layout", default=1, type=int)
        add_server_arguments(p)

        opts = p.parse_args()

        if opts.site == 1:
            site = build_site_node1()
        elif opts.site == 2:
            site = build_site_node2()
        elif opts.site == 3:
            site = build_site_node3()
        else:
            raise p.error("Invalid site value")

        if opts.register_as is None:
            opts.register_as = "node%s" % opts.site

        self.context = await server_context_from_arguments(site, opts)

        if opts.simple:
            rd_wkc = urljoin(opts.rd_uri, '/.well-known/core?ep=%s&lt=6000' % opts.register_as)
            await self.context.request(Message(code=POST, uri=rd_wkc)).response
        else:
            self.registerer = Registerer(self.context, opts.rd_uri, lt=120,
                    registration_parameters={'ep': opts.register_as})

    async def shutdown(self):
        if self.registerer is not None:
            await self.registerer.shutdown()
        await self.context.shutdown()

if __name__ == "__main__":
    RDRegistrant.sync_main()



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""List of known values for the CoAP "Type" field.

As this field is only 2 bits, its valid values are comprehensively enumerated
in the `Type` object.
"""

from enum import IntEnum

class Type(IntEnum):
    CON = 0 # Confirmable
    NON = 1 # Non-confirmable
    ACK = 2 # Acknowledgement
    RST = 3 # Reset

    def __str__(self):
        return self.name

CON, NON, ACK, RST = Type.CON, Type.NON, Type.ACK, Type.RST

__all__ = ['Type', 'CON', 'NON', 'ACK', 'RST']



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""This helper script can be used to easily inspect aiocoap's environment
autodetection (ie. whether all modules required for particular subsystems are
available, losely corresponding to the "features" made available through
setup.py); run it as `python3 -m aiocoap.cli.defaults`."""

import sys
from aiocoap.meta import version
from aiocoap.defaults import dtls_missing_modules, oscore_missing_modules, \
    linkheader_missing_modules, prettyprint_missing_modules, has_reuse_port, \
    get_default_clienttransports, get_default_servertransports
import argparse
import os

if __name__ == "__main__":
    p = argparse.ArgumentParser(description=__doc__)
    # Allow passing this in as AIOCOAP_DEFAULTS_EXPECT_ALL=1 via the
    # environment, as that's easier to set in tox
    p.add_argument("--expect-all", help="Exit with an error unless all subsystems are available",
            action='store_true', default=os.environ.get('AIOCOAP_DEFAULTS_EXPECT_ALL') == '1')
    p.add_argument('--version', action='version', version=version)
    args = p.parse_args()

    error = 0

    print("Python version: %s" % sys.version)
    print("aiocoap version: %s" % version)
    print("Modules missing for subsystems:")
    missmods = [dtls_missing_modules, oscore_missing_modules, linkheader_missing_modules, prettyprint_missing_modules]
    for m in missmods:
        name = m.__name__[:-len("_missing_modules")]
        missing = m()
        if missing and args.expect_all:
            error = 1
        print("    %s: %s" % (name, "everything there" if not missing else "missing " + ", ".join(missing)))
    print("Python platform: %s" % sys.platform)
    print("Default server transports:  %s" % ":".join(get_default_servertransports(use_env=False)))
    print("Selected server transports: %s" % ":".join(get_default_servertransports()))
    print("Default client transports:  %s" % ":".join(get_default_clienttransports(use_env=False)))
    print("Selected client transports: %s" % ":".join(get_default_clienttransports()))
    print("SO_REUSEPORT available (default, selected): %s, %s" % (has_reuse_port(use_env=False), has_reuse_port()))

    if error:
        print("Exiting unsuccessfully because --expect-all was set and not all extras are available.")
    sys.exit(error)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Extensions to asyncio and workarounds around its shortcomings"""

import sys

def py38args(**kwargs):
    """Wrapper around kwargs that replaces them with an empty list for Python
    versions earlier than 3.8.

    This is used to assign a name in asyncio.create_task to pass in a name."""
    if sys.version_info >= (3, 8):
        return kwargs
    else:
        return {}



#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""This is a usage example of aiocoap that demonstrates how to implement a
simple client. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)

async def main():
    protocol = await Context.create_client_context()

    request = Message(code=GET, uri='coap://vs0.inf.ethz.ch/obs', observe=0)

    pr = protocol.request(request)

    r = await pr.response
    print("First response: %s\n%r"%(r, r.payload))

    async for r in pr.observation:
        print("Next result: %s\n%r"%(r, r.payload))

        pr.observation.cancel()
        break
    print("Loop ended, sticking around")
    await asyncio.sleep(50)

if __name__ == "__main__":
    asyncio.run(main())



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

import asyncio
import warnings

from ..message import UndecidedRemote
from .. import interfaces
from ..protocol import ClientObservation

from ..util import hostportsplit
from ..util.asyncio import py38args

class ProxyForwarder(interfaces.RequestProvider):
    """Object that behaves like a Context but only provides the request
    function and forwards all messages to a proxy.

    This is not a proxy itself, it is just the interface for an external
    one."""
    def __init__(self, proxy_address, context):
        if '://' not in proxy_address:
            warnings.warn("Proxy addresses without scheme are deprecated, "
                    "please specify like `coap://host`, `coap+tcp://ip:port` "
                    "etc.", DeprecationWarning)
            proxy_address = 'coap://' + proxy_address

        self.proxy_address = UndecidedRemote.from_pathless_uri(proxy_address)
        self.context = context

    proxy = property(lambda self: self._proxy)

    def request(self, message, **kwargs):
        if not isinstance(message.remote, UndecidedRemote):
            raise ValueError(
                "Message already has a configured "\
                "remote, set .opt.uri_{host,port} instead of remote")
        host, port = hostportsplit(message.remote.hostinfo)
        message.opt.uri_port = port
        message.opt.uri_host = host
        message.opt.proxy_scheme = self.proxy_address.scheme
        message.remote = self.proxy_address

        return self.context.request(message)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Module in which all meaningful numbers are collected. Most of the submodules
correspond to IANA registries.

The contents of the :mod:`.constants`, :mod:`.types` and :mod:`.codes` modules
are accessible through this module directly; :mod:`.contentformat`'s and
:mod:`.optionnumbers`' sole :class:`.contentformat.ContentFormat` and
:class:`.optionnumbers.OptionNumber` classes are accessible in the same way.
"""

from . import constants, types, codes
# flake8 doesn't see through the global re-export
from .constants import * # noqa: F401 F403
from .types import * # noqa: F401 F403
from .codes import * # noqa: F401 F403
from .contentformat import ContentFormat, _MediaTypes, _MediaTypesRev
from .optionnumbers import OptionNumber

__all__ = constants.__all__ + types.__all__ + codes.__all__ + ['OptionNumber', 'ContentFormat']

media_types = _MediaTypes()
media_types_rev = _MediaTypesRev()



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Implementations of draft-ietf-core-resource-directory

Currently, this only contains the :mod:`client` module. There is an
implementation of the server side contained in aiocoap, but it is currently
accessible only as :mod:`command line application <aiocoap.cli.rd>`.
"""



#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

import logging
from aiocoap import oscore_sitewrapper
import aiocoap.resource as resource
import aiocoap
import asyncio

logging.basicConfig(level=logging.INFO)
logging.getLogger("coap-server").setLevel(logging.DEBUG)

async def main():
    # Resource tree creation
    root = resource.Site()

    root.add_resource(['.well-known', 'core'],
            resource.WKCResource(root.get_resources_as_linkheader))

    server_credentials = aiocoap.credentials.CredentialsMap()

    root = oscore_sitewrapper.OscoreSiteWrapper(root, server_credentials)

    protocol = await aiocoap.Context.create_server_context(root)

    # Keys from IETF109 plug test: Rikard Test 2 Entity 1
    server_credentials[":a"] = \
            aiocoap.oscore.SimpleGroupContext(
                    algorithm = aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
                    hashfun = aiocoap.oscore.hashfunctions[aiocoap.oscore.DEFAULT_HASHFUNCTION],
                    alg_countersign = aiocoap.oscore.Ed25519(),
                    group_id = bytes.fromhex('DD11'),
                    master_secret = bytes.fromhex('11223344556677889900AABBCCDDEEFF'),
                    master_salt = bytes.fromhex('1F2E3D4C5B6A7081'),
                    sender_id = bytes.fromhex('0A'),
                    private_key = bytes.fromhex('397CEB5A8D21D74A9258C20C33FC45AB152B02CF479B2E3081285F77454CF347'),
                    peers = {
                        bytes.fromhex('51'): bytes.fromhex('2668BA6CA302F14E952228DA1250A890C143FDBA4DAED27246188B9E42C94B6D'),
                        bytes.fromhex('52'): bytes.fromhex('5394E43633CDAC96F05120EA9F21307C9355A1B66B60A834B53E9BF60B1FB7DF'),
                        bytes.fromhex('dc'): aiocoap.oscore.DETERMINISTIC_KEY,
                        },
                    )

    await asyncio.get_running_loop().create_future()

if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.
import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)

async def main():
    protocol = await Context.create_client_context()

    import aiocoap.oscore
    # Acting as Rikard Test 2 Entity 3
    protocol.client_credentials["coap://localhost/*"] = \
            aiocoap.oscore.SimpleGroupContext(
                    algorithm = aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
                    hashfun = aiocoap.oscore.hashfunctions[aiocoap.oscore.DEFAULT_HASHFUNCTION],
                    alg_countersign = aiocoap.oscore.Ed25519(),
                    group_id = bytes.fromhex('DD11'),
                    master_secret = bytes.fromhex('11223344556677889900AABBCCDDEEFF'),
                    master_salt = bytes.fromhex('1F2E3D4C5B6A7081'),
                    # these are really moot given we only access this as a deterministic client
                    sender_id = bytes.fromhex('52'),
                    private_key = bytes.fromhex('E550CD532B881D52AD75CE7B91171063E568F2531FBDFB32EE01D1910BCF810F'),

                    peers = {
                        bytes.fromhex('0A'): bytes.fromhex('CE616F28426EF24EDB51DBCEF7A23305F886F657959D4DF889DDFC0255042159'),
                        bytes.fromhex('51'): bytes.fromhex('2668BA6CA302F14E952228DA1250A890C143FDBA4DAED27246188B9E42C94B6D'),
                        # that's a new one for deterministic, and we build a key for it but no shared secret
                        bytes.fromhex('dc'): None,
                        },
# requests in group mode would be sent with
#                    ).pairwise_for(bytes.fromhex('0A'))
# but here we want to request deterministically
                    ).for_sending_deterministic_requests(bytes.fromhex("dc"), bytes.fromhex('0a'))

    request = Message(code=GET, uri='coap://localhost/.well-known/core')
    response = await protocol.request(request).response
    print("Got response", response, response.payload)

    request = Message(code=GET, uri='coap://localhost/.well-known/core')
    response = await protocol.request(request).response
    print("Got response", response, response.payload)

if __name__ == "__main__":
    asyncio.run(main())



import functools
import re
import os
import textwrap
import glob
import os.path

def glob35(dir, recursive=True):
    return glob.glob(dir.replace('**', '*')) + \
            glob.glob(dir.replace('**', '*/*')) + \
            glob.glob(dir.replace('**', '*/*/*')) + \
            glob.glob(dir.replace('**', '*/*/*/*')) + \
            glob.glob(dir.replace('/**', ''))

from docutils.parsers.rst.directives.misc import Include

rtd_re = re.compile("^\\.\\. _([^:]+): http://aiocoap.readthedocs.io/en/latest/(.*)\\.html$")
filelink_re = re.compile("^\\.\\. _([^:]+): ([^:]+)$")

def addrepl(replacements, pattern, prefix, linkbody):
    """Given details of a sphinx-ified link, place an entry in the replacements
    dictionary that is suitable for full-text substitution later on"""
    pretty = pattern.strip('`')
    if pretty != linkbody:
        sphinxlink = ':%s:`%s <%s>`'%(prefix, pretty, linkbody)
    else:
        sphinxlink = ':%s:`%s`'%(prefix, linkbody)
    replacements[pattern + '_'] = sphinxlink

def modified_insert_input(include_lines, path, original=None):
    """A filter for the insert_input function that preprocesses the input in
    the following ways:

    * Remove all external hyperlink targets to readthedocs (and guess the way
      this link needs to be expressed in Sphinx)
    * Remove all local (relative) link targets
    * Replace words referencing them with the appropriate Sphinx reference"""
    new_lines = []
    replacements = {}
    for line in include_lines:
        rtd_match = rtd_re.match(line)
        filelink_match = filelink_re.match(line)
        if rtd_match:
            pattern, linkbody = rtd_match.groups()
            prefix = 'doc'
            addrepl(replacements, pattern, prefix, linkbody)
        elif filelink_match:
            # for things like LICENSE
            pattern, linkbody = filelink_match.groups()
            addrepl(replacements, pattern, 'doc', linkbody)
        else:
            # only keep lines that are still relevant (that's most lines)
            new_lines.append(line)
    new_lines = [functools.reduce(lambda s, rep: s.replace(*rep), replacements.items(), x) for x in new_lines]
    original(new_lines, path)

class IncludePreprocessed(Include):
    """Include a file (like the 'Include' directive), but preprocess its input
    as described in modified_insert_input."""
    def run(self):
        self.state_machine.insert_input = functools.partial(modified_insert_input, original=self.state_machine.insert_input)
        try:
            result = super().run()
        finally:
            del self.state_machine.insert_input
        return result

def build_moduledocs(app):
    """Create per-module sources like sphinx-apidoc, but at build time and with
    customizations."""
    srcdir = app.builder.srcdir

    moddir = srcdir + '/module'
    os.makedirs(moddir, exist_ok=True)

    basedir = os.path.dirname(srcdir)
    docs = [x[len(basedir)+1:-3].replace('/', '.').replace('.__init__', '') for x in glob35(basedir + '/aiocoap/**/*.py', recursive=True)]

    for x in docs:
        commonstart = textwrap.dedent("""\
            {x} module
            ====================================================================================
            """).format(x=x)

        if x in ('aiocoap.numbers', 'aiocoap.transports'):
            # They have explicit intros pointing out submodules and/or
            # describing any reexports
            text = commonstart + textwrap.dedent("""
                .. automodule:: {x}
                .. toctree::
                    :glob:

                    {x}.*
                """).format(x=x)
        elif x.startswith('aiocoap.cli.'):
            executablename = "aiocoap-" + x[len('aiocoap.cli.'):]
            # no ".. automodule:: {x}" because the doc string is already used
            # by the argparse, and thus would be repeated
            text = textwrap.dedent("""
                    {executablename}
                    ==============================

                    .. argparse::
                        :ref: {x}.build_parser
                        :prog: {executablename}

                    """).format(x=x, executablename=executablename)
        else:
            text = commonstart + textwrap.dedent("""
                .. automodule:: {x}
                    :members:
                    :undoc-members:
                    :show-inheritance:
                """).format(x=x)
        docname = "%s/%s.rst"%(moddir, x)

        if os.path.exists(docname) and open(docname).read() == text:
            continue
        else:
            with open(moddir + '/' + x + '.rst', 'w') as outfile:
                outfile.write(text)

    for f in os.listdir(moddir):
        if f.endswith('.rst') and f[:-4] not in docs:
            os.unlink(moddir + '/' + f)

def setup(app):
    """Sphinx extension that builds the aiocoap index page from a non-sphinx
    and thus github-suitable ReST file, and also creates sphinx-apidoc style
    per-module pages customized to what I'd like to see there"""

    app.add_directive('include_preprocessed', IncludePreprocessed)
    app.connect('builder-inited', build_moduledocs)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Constants either defined in the CoAP protocol (often default values for lack
of ways to determine eg. the estimated round trip time). Some parameters are
invented here for practical purposes of the implementation (eg.
DEFAULT_BLOCK_SIZE_EXP, EMPTY_ACK_DELAY)."""

COAP_PORT = 5683
"""The IANA-assigned standard port for COAP services."""

COAPS_PORT = 5684

MCAST_IPV4_ALLCOAPNODES = "224.0.1.187"
MCAST_IPV6_LINKLOCAL_ALLNODES = "ff02::1"
MCAST_IPV6_LINKLOCAL_ALLCOAPNODES = "ff02::fd"
MCAST_IPV6_SITELOCAL_ALLNODES = "ff05::1"
MCAST_IPV6_SITELOCAL_ALLCOAPNODES = "ff05::fd"
MCAST_ALL = (
      MCAST_IPV4_ALLCOAPNODES,
      MCAST_IPV6_LINKLOCAL_ALLNODES,
      MCAST_IPV6_LINKLOCAL_ALLCOAPNODES,
      MCAST_IPV6_SITELOCAL_ALLNODES,
      MCAST_IPV6_SITELOCAL_ALLCOAPNODES,
      )

#   +-------------------+---------------+
#   | name              | default value |
#   +-------------------+---------------+
#   | ACK_TIMEOUT       | 2 seconds     |
#   | ACK_RANDOM_FACTOR | 1.5           |
#   | MAX_RETRANSMIT    | 4             |
#   | NSTART            | 1             |
#   | DEFAULT_LEISURE   | 5 seconds     |
#   | PROBING_RATE      | 1 Byte/second |
#   +-------------------+---------------+

ACK_TIMEOUT = 2.0
"""The time, in seconds, to wait for an acknowledgement of a
confirmable message. The inter-transmission time doubles
for each retransmission."""

ACK_RANDOM_FACTOR = 1.5
"""Timeout multiplier for anti-synchronization."""

MAX_RETRANSMIT = 4
"""The number of retransmissions of confirmable messages to
non-multicast endpoints before the infrastructure assumes no
acknowledgement will be received."""

NSTART = 1
"""Maximum number of simultaneous outstanding interactions
   that endpoint maintains to a given server (including proxies)"""

#   +-------------------+---------------+
#   | name              | default value |
#   +-------------------+---------------+
#   | MAX_TRANSMIT_SPAN |          45 s |
#   | MAX_TRANSMIT_WAIT |          93 s |
#   | MAX_LATENCY       |         100 s |
#   | PROCESSING_DELAY  |           2 s |
#   | MAX_RTT           |         202 s |
#   | EXCHANGE_LIFETIME |         247 s |
#   | NON_LIFETIME      |         145 s |
#   +-------------------+---------------+

MAX_TRANSMIT_SPAN = ACK_TIMEOUT * (2 ** MAX_RETRANSMIT - 1) * ACK_RANDOM_FACTOR
"""Maximum time from the first transmission
of a confirmable message to its last retransmission."""

MAX_TRANSMIT_WAIT = ACK_TIMEOUT * (2 ** (MAX_RETRANSMIT + 1) - 1) * ACK_RANDOM_FACTOR
"""Maximum time from the first transmission
of a confirmable message to the time when the sender gives up on
receiving an acknowledgement or reset."""

MAX_LATENCY = 100.0
"""Maximum time a datagram is expected to take from the start
of its transmission to the completion of its reception."""

PROCESSING_DELAY = ACK_TIMEOUT
""""Time a node takes to turn around a
confirmable message into an acknowledgement."""

MAX_RTT = 2 * MAX_LATENCY + PROCESSING_DELAY
"""Maximum round-trip time."""

EXCHANGE_LIFETIME = MAX_TRANSMIT_SPAN + MAX_RTT
"""time from starting to send a confirmable message to the time when an
acknowledgement is no longer expected, i.e. message layer information about the
message exchange can be purged"""

DEFAULT_BLOCK_SIZE_EXP = 6 # maximum block size 1024
"""Default size exponent for blockwise transfers."""

EMPTY_ACK_DELAY = 0.1
"""After this time protocol sends empty ACK, and separate response"""

REQUEST_TIMEOUT = MAX_TRANSMIT_WAIT
"""Time after which server assumes it won't receive any answer.
   It is not defined by IETF documents.
   For human-operated devices it might be preferable to set some small value
   (for example 10 seconds)
   For M2M it's application dependent."""

DEFAULT_LEISURE = 5

MULTICAST_REQUEST_TIMEOUT = REQUEST_TIMEOUT + DEFAULT_LEISURE

OBSERVATION_RESET_TIME = 128
"""Time in seconds after which the value of the observe field are ignored.

This number is not explicitly named in RFC7641.
"""

SHUTDOWN_TIMEOUT = 3
"""Maximum time, in seconds, for which the process is kept around during shutdown"""

__all__ = [k for k in dir() if not k.startswith('_')]



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Container module for command line utilities bundled with aiocoap.

These modules are not considered to be a part of the aioCoAP API, and are thus
subject to change even when the project reaches a stable version number. If you
want to use any of that infrastructure, please file a feature request for
stabilization in the project's issue tracker.

The tools themselves are documented in :doc:`/tools`.
"""



#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""This is a usage example of aiocoap that demonstrates how to implement a
simple client. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)

async def main():
    protocol = await Context.create_client_context()

    request = Message(code=GET, uri='coap://localhost/time')

    try:
        response = await protocol.request(request).response
    except Exception as e:
        print('Failed to fetch resource:')
        print(e)
    else:
        print('Result: %s\n%r'%(response.code, response.payload))

if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3

# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""aiocoap is a Python library for writing servers and clients for the CoAP
(Constrained Application) Protocol, which is used mainly in the context of IoT
(Internet of Things) devices."""

from setuptools import setup, find_packages, Command
import os
import os.path

name = "aiocoap"
version = "0.4.5" # Don't forget meta.version and doc/conf.py
description = "Python CoAP library"

# When introducing something new, make sure to update doc/installation.rst
extras_require = {
        'linkheader': ['LinkHeader'],
        # ge25519 is a workaround for
        # <https://github.com/pyca/cryptography/issues/5557>; being pure python
        # it's light enough to not warrant a dedicated group-oscore extra.
        'oscore': ['cbor2', 'cryptography (>= 2.0)', 'filelock', 'ge25519'],
        'tinydtls': ['DTLSSocket >= 0.1.11a1'],
        'ws': ['websockets'],
        'prettyprint': ['termcolor', 'cbor2', 'LinkHeader', 'pygments'],
        'docs': ['sphinx', 'sphinx-argparse'], # extended below
        'all': [], # populated below, contains everything but documentation dependencies for easier installation
        }
tests_require = [] # populated below

test_extras = extras_require.keys()
if 'AIOCOAP_TEST_EXTRAS' in os.environ:
    test_extras = os.environ['AIOCOAP_TEST_EXTRAS'].split(':')

for k, v in extras_require.items():
    if k.startswith(':') or k == 'all' or k == 'docs':
        continue

    # Most extras are required for docs to build. TinyDTLS is an exception
    # because no module imports it at module level. (This is convenient also
    # because TinyDTLS installation fails on readthedocs for unknown reasons.)
    if k != 'tinydtls':
        extras_require['docs'].extend(v)

    extras_require['all'].extend(v)
    if k in test_extras:
        tests_require.extend(v)

class Cite(Command):
    description = """Print how to cite aiocoap in a publication"""

    user_options = [("bibtex", None, "Output citation data as bibtex")]
    boolean_options = ["bibtex"]

    def initialize_options(self):
        self.bibtex = False

    def finalize_options(self):
        pass

    def run(self):
        if self.bibtex:
            print(self.bibtex_text)
        else:
            print(self.plain_text)

    plain_text = """Amsüss, Christian and Wasilak, Maciej. aiocoap: Python CoAP Library. Energy Harvesting Solutions, 2013–. http://github.com/chrysn/aiocoap/"""

    bibtex_text = """@Misc{,
        author = {Christian Amsüss and Maciej Wasilak},
        organization = {Energy Harvesting Solutions},
        title = {{aiocoap}: Python CoAP Library},
        year = {2013--},
        url = {http://github.com/chrysn/aiocoap/},
        }"""

setup(
    name=name,
    version=version,
    description=description,
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=["tests"]),

    author="Maciej Wasilak, Christian Amsüss",
    author_email="c.amsuess@energyharvesting.at",
    url="https://github.com/chrysn/aiocoap",

    license='MIT',
    keywords=['coap', 'asyncio', 'iot'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
        ],

    # When changing this, also look into doc/faq.rst and README.rst
    python_requires='>=3.7',
    extras_require=extras_require,
    tests_require=tests_require,

    # see doc/README.doc seciton "dependency hack"
    install_requires=extras_require['docs'] if 'READTHEDOCS' in os.environ else [],

    entry_points={
        'console_scripts': [
            'aiocoap-client = aiocoap.cli.client:sync_main',
            'aiocoap-proxy = aiocoap.cli.proxy:sync_main',
            'aiocoap-rd = aiocoap.cli.rd:sync_main [linkheader]',
            # link header is no important dependency (only used for RD
            # registration), but currently a hard one
            'aiocoap-fileserver = aiocoap.cli.fileserver:FileServerProgram.sync_main [linkheader]',
            ]
        },

    cmdclass={
        'cite': Cite,
        },

    # not strictly required any more since tests are now runnable as `-m
    # unittest`, but results in more concise output
    test_suite='tests',
)



import time

class _Throttler:
    """Wrapper around an argumentless function that silently drops calls if
    there are too many."""
    # FIXME i'd rather have the ObservableResource or even the observation
    # itself handle this
    def __init__(self, callback):
        self.callback = callback
        self.last = 0

    def __call__(self):
        now = time.time()
        if now - self.last < 0.2:
            return
        self.last = now
        self.callback()



import asyncio

from .resource import *
from .util import _Throttler

class KivyPropertyBacked(SenmlResource):
    """Provides a .value bound to the self.widget_property property of the
    self.backend_widget Kivy widget. The widget and widget_property need to be
    set before this class's __init__ is called."""

    def __init__(self):
        super().__init__()
        throttler = _Throttler(self.value_changed)
        asyncio.get_event_loop().create_task(self._monitor(throttler))

    async def _monitor(self, throttler):
        async for evt in self.backend_widget.async_bind(self.widget_property):
            throttler()

    def _get_value(self):
        return getattr(self.backend_widget, self.widget_property)

    def _set_value(self, updated):
        setattr(self.backend_widget, self.widget_property, updated)

    value = property(_get_value, _set_value)

class Text(StringResource, KivyPropertyBacked):
    """A resource that represents a Kivy widget's text"""

    if_ = "core.p"

    widget_property = 'text'

    def __init__(self, kivy_backend):
        self.backend_widget = kivy_backend
        super().__init__()

class PressState(BooleanResource, KivyPropertyBacked):
    """A resource that resesents a Kivy widget's "pressed" property"""

    if_ = "core.s"

    widget_property = 'state'

    # This is no good modelling (for a button modelled as press-release is hard
    # to observe with eventualy consistent semantics only), but it works for a
    # first demo.

    def __init__(self, kivy_backend):
        self.backend_widget = kivy_backend
        super().__init__()

    # fake boolean, and don't allow setting

    def _get_value(self):
        return {'down': 1, 'normal': 0}[super()._get_value()]

    value = property(_get_value)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Container module, see submodules:

* :mod:`.client` -- using CoAP via a proxy server
* :mod:`.server` -- running a proxy server
"""



# This requires Python 3.6 as it makes use of __init_subclass__

from collections import namedtuple, defaultdict
import functools
import inspect
import abc

import json
import cbor2 as cbor

from aiocoap import resource, numbers, interfaces
from aiocoap import GET, PUT, POST, Message, CONTENT, CHANGED
from aiocoap.error import BadRequest, UnsupportedContentFormat, UnallowedMethod


_ContenttypeRenderedHandler = namedtuple("_ContenttypeRenderedHandler", ("method", "accept", "contentformat", "implementation", "responseformat"))

# this could become an alternative to the resource.Resource currently implemented in aiocoap.resource

class ContenttypeRendered(resource._ExposesWellknownAttributes, interfaces.Resource, metaclass=abc.ABCMeta):
    def __init_subclass__(cls):
        # __new__ code moved in here to use __ properties
        cls.__handlers = defaultdict(lambda: {})
        for member in vars(cls).values():
            if isinstance(member, _ContenttypeRenderedHandler):
                for accept in member.accept:
                    for contentformat in member.contentformat:
                        cls.__handlers[member.method][(accept, contentformat)] = (member.implementation, member.responseformat)

    @staticmethod
    def get_handler(accept, *, default=False):
        """Decorate a method with this to make it the GET handler for a given
        method and Accept value (or additionally the empty one if default=True).

        FIXME move to ContenttypeRendered

        Methods that accept a payload will get the payload passed as an
        argument, and be decided based on the Content-Format header (with the
        Accept header being ignored; possibly, that's a reasons to split this
        decorator up per-method).

        The method will not be usable by its name any more. It is recommended
        to use a double-underscore name for thusly decorated methods (eg.
        __get_plain).

        The method has some freedom in the types it may return (None is treated
        as an empty payload, strings are encoded in UTF-8). It is unclear yet
        whether more complex conversions (eg. JSON, CBOR) will be supported by
        this or need additonal decorators."""
        def wrapper(func):
            cf = numbers.media_types_rev[accept]
            return _ContenttypeRenderedHandler(GET, (cf, None) if default else (cf,), (None,), func, cf)
        return wrapper

    @staticmethod
    def put_handler(contentformat, *, default=False):
        def wrapper(func):
            cf = numbers.media_types_rev[contentformat]
            return _ContenttypeRenderedHandler(PUT, (None,), (cf, None) if default else (cf,), func, None)
        return wrapper

    @staticmethod
    def empty_post_handler():
        # i suppose this'll be replaced with something more generic when i add something that needs request or response payloads
        def wrapper(func):
            return _ContenttypeRenderedHandler(POST, (None,), (None,), func, None)
        return wrapper

    async def needs_blockwise_assembly(self, request):
        return True

    async def render(self, request):
        cf = request.opt.content_format
        acc = request.opt.accept
        raise_class = UnallowedMethod
        method_would_have_worked = False

        # FIXME: manually walking the MRO is not a nice way to go about this;
        # is there no other way to query the registered handlers according to
        # the regular inheritance patterns?
        for cls in type(self).mro():
            if not issubclass(cls, ContenttypeRendered) or cls is ContenttypeRendered:
                continue
            for_method = cls.__handlers.get(request.code, None)
            if for_method is None:
                continue
            raise_class = UnsupportedContentFormat
            handler, responseformat = for_method.get((acc, cf), (None, None))
            if handler is not None:
                break
        else:
            raise raise_class()

        sig = inspect.signature(handler)
        parameters = set(sig.parameters.keys())
        parameters.remove("self")
        kwargs = {}
        if request.payload and "payload" not in parameters:
            raise BadRequest("Unexpected payload")
        if request.opt.uri_query and "query" not in parameters:
            raise BadRequest("Unexepcted query arguments")

        for p in parameters:
            if p == "payload":
                kwargs['payload'] = request.payload
            elif p == "request_uri":
                # BIG FIXME: This does not give the expected results due to the
                # URI path stripping in Site, and because Message gets the
                # requested authority wrong on the server side.
                kwargs["request_uri"] = request.get_request_uri()
            else:
                raise RuntimeError("Unexpected argument requested: %s" % p)
        payload = handler(self, **kwargs)

        if payload is None:
            payload = b""
        elif isinstance(payload, str):
            payload = payload.encode('utf8')

        return Message(
                code={GET: CONTENT, PUT: CHANGED}[request.code],
                payload=payload,
                content_format=responseformat,
                no_response=request.opt.no_response,
                )

class ObservableContenttypeRendered(ContenttypeRendered, interfaces.ObservableResource):
    def __init__(self):
        super().__init__()

        self._callbacks = set()

    async def add_observation(self, request, serverobservation):
        """Implementation of interfaces.ObservableResource"""
        callback = serverobservation.trigger
        self._callbacks.add(callback)
        remover = functools.partial(self._callbacks.remove, callback)
        serverobservation.accept(remover)

    def add_valuechange_callback(self, cb):
        """Call this when you want a callback outside of aiocoap called
        whenever value_change is called, typically because the callback
        recipient would extract the state of the resource in a non-CoAP way."""
        self._callbacks.add(cb)

    def value_changed(self):
        """Call this whenever the object was modified in such a way that any
        rendition might change."""
        for c in self._callbacks:
            c()


class SenmlResource(ObservableContenttypeRendered):
    """A resource that has its state in .value; this class implements SenML
    getters and setters as well as plain text.

    Implementors need to provide a .value instance property as well as
    .jsonsenml_key / .cborsenml_key class properties for picking the right
    value key in the respective SenML serialization, and a .valuetype type that
    is used both for converting any text/plain'ly PUT string as well as for
    filtering (typically copy-constructing) data from SenML."""

    @ContenttypeRendered.get_handler('application/senml+json')
    def __jsonsenml_get(self, request_uri):
        return json.dumps([{"n": request_uri, self.jsonsenml_key: self.value}])

    @ContenttypeRendered.get_handler('application/senml+cbor')
    def __cborsenml_get(self, request_uri):
        return cbor.dumps([{0: request_uri, self.cborsenml_key: self.value}])

    @ContenttypeRendered.get_handler('text/plain;charset=utf-8', default=True)
    def __textplain_get(self):
        return str(self.value)

    @ContenttypeRendered.put_handler('application/senml+json')
    def __jsonsenml_set(self, payload, request_uri):
        try:
            new = json.loads(payload.decode('utf8'))
            if len(new) != 1 or new[0].get("bn", "") + new[0].get("n", "") != request_uri:
                raise BadRequest("Not a single record pertaining to this resource")
            self.value = self.valuetype(new[0][self.jsonsenml_key])
        except (KeyError, ValueError):
            raise BadRequest()

    @ContenttypeRendered.put_handler('application/senml+cbor')
    def __cborsenml_set(self, payload, request_uri):
        try:
            new = cbor.loads(payload)
            if len(new) != 1 or new[0].get(-2, "") + new[0].get(0, "") != request_uri:
                raise BadRequest("Not a single record pertaining to this resource")
            self.value = self.valuetype(new[self.cborsenml_key])
        except (KeyError, ValueError):
            raise BadRequest()

    @ContenttypeRendered.put_handler('text/plain;charset=utf-8', default=True)
    def __textplain_set(self, payload):
        try:
            self.value = self.valuetype(payload.decode('utf8').strip())
        except ValueError:
            raise BadRequest()

class BooleanResource(SenmlResource):
    jsonsenml_key = "vb"
    cborsenml_key = 4
    valuetype = bool

    @ContenttypeRendered.get_handler('text/plain;charset=utf-8', default=True)
    def __textplain_get(self):
        return "01"[self.value]

    @ContenttypeRendered.put_handler('text/plain;charset=utf-8', default=True)
    def __textplain_set(self, payload):
        try:
            self.value = {"0": False, "1": True}[payload.decode('utf8').strip()]
        except (KeyError, ValueError):
            raise BadRequest()

class FloatResource(SenmlResource):
    jsonsenml_key = "v"
    cborsenml_key = 2
    valuetype = float

class StringResource(SenmlResource):
    jsonsenml_key = "vs"
    cborsenml_key = 3
    valuetype = str

class SubsiteBatch(ObservableContenttypeRendered):
    """An implementation of a CoRE interfaces batch that is the root resource
    of a subsite

    This currently depends on being added to the site after all other
    resources; it could enumerate them later, but it installs its own
    value_changed callbacks of other members at initialization time."""

    if_ = 'core.b'

    def __init__(self, site):
        self.site = site
        super().__init__()

        # FIXME this ties in directly into resource.Site's privates, AND it
        # should actually react to changes in the site, AND overriding the
        # callback to install an own hook is not compatible with any other
        # ObservableResource implementations
        for subres in self.site._resources.values():
            if isinstance(subres, ObservableContenttypeRendered):
                subres.add_valuechange_callback(self.value_changed)
        for subsite in self.site._subsites.values():
            if not isinstance(subsite, resource.Site):
                continue # can't access privates
            if () not in subsite._resources:
                continue # no root, better not try
            rootres = subsite._resources[()]
            if not isinstance(rootres, SubsiteBatch):
                continue
            rootres.add_valuechange_callback(self.value_changed)

    def __get_records(self, request_uri):
        records = []
        # FIXME this ties in directly into resource.Site's privates
        for path, subres in self.site._resources.items():
            if isinstance(subres, SenmlResource): # this conveniently filters out self as well
                records.append({'n': '/'.join(path), subres.jsonsenml_key: subres.value})
        print(self.site, vars(self.site))
        for path, subsite in self.site._subsites.items():
            if not isinstance(subsite, resource.Site):
                continue # can't access privates
            if () not in subsite._resources:
                continue # no root, better not try
            rootres = subsite._resources[()]
            if not isinstance(rootres, SubsiteBatch):
                continue
            for r in rootres.__get_records(request_uri):
                r = dict(**r)
                r.pop('bn', None)
                r['n'] = "/".join(path) + "/" + r['n']
                records.append(r)
        records[0]['bn'] = request_uri
        return records

    @ContenttypeRendered.get_handler('application/senml+json', default=True)
    def __regular_get(self, request_uri):
        return json.dumps(self.__get_records(request_uri))


class PythonBacked(SenmlResource):
    """Provides a .value stored in regular Python, but pulls the
    .value_changed() trigger on every change"""

    def _set_value(self, value):
        changed = not hasattr(self, '_value') or self._value != value
        self._value = value
        if changed:
            self.value_changed()

    value = property(lambda self: self._value, _set_value)
